"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createNoUnsafeAssignmentRule = void 0;
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
const typescript_1 = require("typescript");
const tsutils_1 = require("tsutils");
const isSignatureAssignable = (destinationNode, sourceNode, destinationSignature, sourceSignature, checker) => {
    const returnTypeIsAssignable = checker.isTypeAssignableTo(sourceSignature.getReturnType(), destinationSignature.getReturnType());
    const allParametersAreAssignable = destinationSignature.parameters.every((destinationParameter, index) => {
        const sourceParameter = sourceSignature.parameters[index];
        // eslint-disable-next-line functional/no-conditional-statement
        if (sourceParameter === undefined) {
            return false;
        }
        // source and destination parameter reversed here due to contravariance.
        return checker.isTypeAssignableTo(checker.getTypeOfSymbolAtLocation(destinationParameter, destinationNode), checker.getTypeOfSymbolAtLocation(sourceParameter, sourceNode));
    });
    return returnTypeIsAssignable && allParametersAreAssignable;
};
const assignableSignaturePairs = (destinationNode, sourceNode, destinationSignatures, sourceSignatures, checker
// TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
// eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
) => {
    // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
    // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
    return sourceSignatures.flatMap((sourceSignature) => destinationSignatures
        .filter((destinationSignature) => isSignatureAssignable(destinationNode, sourceNode, sourceSignature, destinationSignature, checker))
        .map((destinationSignature) => ({
        sourceSignature,
        destinationSignature,
    })));
};
/**
 * Breaks the supplied types into their union type parts and returns an
 * array of pairs of constituent types that are assignable.
 */
const assignableTypePairs = (rawDestinationType, rawSourceType, checker
// TODO remove this eslint-disable
// eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
) => {
    // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
    // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
    const destinationTypeParts = (0, tsutils_1.unionTypeParts)(rawDestinationType);
    // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
    // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
    const sourceTypeParts = (0, tsutils_1.unionTypeParts)(rawSourceType);
    // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
    // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
    return sourceTypeParts.flatMap((sourceTypePart) => destinationTypeParts
        .filter((destinationTypePart) => checker.isTypeAssignableTo(sourceTypePart, destinationTypePart))
        .map((destinationTypePart) => ({
        sourceType: sourceTypePart,
        destinationType: destinationTypePart,
    })));
};
const createNoUnsafeAssignmentRule = (unsafePropertyAssignmentFunc, unsafeIndexAssignmentFunc) => 
// eslint-disable-next-line sonarjs/cognitive-complexity
(context) => {
    const parserServices = experimental_utils_1.ESLintUtils.getParserServices(context);
    const checker = parserServices.program.getTypeChecker();
    const isUnsafeIndexAssignment = (indexKind, destinationNode, sourceNode, destinationType, sourceType, checker, seenTypes) => {
        const isUnsafe = unsafeIndexAssignmentFunc(indexKind, destinationType, sourceType, checker);
        const destinationIndexType = indexKind === typescript_1.IndexKind.Number
            ? destinationType.getNumberIndexType()
            : destinationType.getStringIndexType();
        const sourceIndexType = indexKind === typescript_1.IndexKind.Number
            ? sourceType.getNumberIndexType()
            : sourceType.getStringIndexType();
        // This is unsafe if...
        return (
        // this particular rule considers the index assignment unsafe, or
        isUnsafe ||
            // the index types are considered unsafe themselves (recursively).
            (destinationIndexType !== undefined &&
                sourceIndexType !== undefined &&
                isUnsafeAssignment(destinationNode, sourceNode, destinationIndexType, sourceIndexType, checker, seenTypes)));
    };
    const isUnsafePropertyAssignmentRec = (destinationNode, sourceNode, 
    // eslint-disable-next-line @typescript-eslint/ban-types
    destinationProperty, 
    // eslint-disable-next-line @typescript-eslint/ban-types
    sourceProperty, checker, seenTypes) => {
        const destinationPropertyType = checker.getTypeOfSymbolAtLocation(destinationProperty, destinationNode);
        const sourcePropertyType = checker.getTypeOfSymbolAtLocation(sourceProperty, sourceNode);
        return isUnsafeAssignment(destinationNode, sourceNode, destinationPropertyType, sourcePropertyType, checker, seenTypes);
    };
    const isUnsafePropertyAssignment = (destinationNode, sourceNode, destinationType, sourceType, checker, seenTypes) => {
        // eslint-disable-next-line functional/no-conditional-statement
        if (checker.isArrayType(destinationType) &&
            !(0, tsutils_1.isUnionOrIntersectionType)(destinationType) &&
            !(0, tsutils_1.isUnionOrIntersectionType)(sourceType)) {
            // For performance reasons we avoid checking every array property for unsafe assignment.
            //
            // If one of these arrays is readonly and the other mutable, it will be caught by `isUnsafeIndexAssignment`,
            // so we needn't compare each individual property.
            //
            // Additionally, the length property of tuples is technically mutable (wtf) so we ignore arrays for this reason too.
            // Observe:
            //   const foo = [] as const;
            //   foo.length = 0;
            //
            // TODO what about tuples?
            // TODO should we ignore other built-in types like strings, symbols and numbers for performance too?
            return false;
        }
        return destinationType.getProperties().some((destinationProperty) => {
            const sourceProperty = sourceType.getProperty(destinationProperty.name);
            const isUnsafe = unsafePropertyAssignmentFunc(destinationProperty, sourceProperty, destinationType, sourceType, checker);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafe) {
                return true;
            }
            // eslint-disable-next-line functional/no-conditional-statement
            if (sourceProperty === undefined) {
                return false;
            }
            // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            const nextSeenTypes = seenTypes.concat({
                destinationType,
                sourceType,
            });
            return isUnsafePropertyAssignmentRec(destinationNode, sourceNode, destinationProperty, sourceProperty, checker, nextSeenTypes);
        });
    };
    const isUnsafeAssignment = (destinationNode, sourceNode, rawDestinationType, rawSourceType, checker, seenTypes = []) => {
        // eslint-disable-next-line functional/no-conditional-statement
        if (rawDestinationType === rawSourceType) {
            // Never unsafe if the types are equal.
            return false;
        }
        // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
        // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
        const nextSeenTypes = seenTypes.concat({
            destinationType: rawDestinationType,
            sourceType: rawSourceType,
        });
        const typePairs = assignableTypePairs(rawDestinationType, rawSourceType, checker);
        // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
        // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
        const objectTypePairs = typePairs.filter((tp) => (0, tsutils_1.intersectionTypeParts)(tp.destinationType).some(tsutils_1.isObjectType));
        const isUnsafeFunctionAssignment = (typePairs) => typePairs.some(({ sourceType, destinationType }) => {
            // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            const nextSeenTypesWithPair = nextSeenTypes.concat({
                destinationType,
                sourceType,
            });
            const sourceCallSignatures = (0, tsutils_1.getCallSignaturesOfType)(sourceType);
            const destinationCallSignatures = (0, tsutils_1.getCallSignaturesOfType)(destinationType);
            const signaturePairs = assignableSignaturePairs(destinationNode, sourceNode, destinationCallSignatures, sourceCallSignatures, checker);
            return signaturePairs.some(({ destinationSignature, sourceSignature }) => {
                // eslint-disable-next-line functional/functional-parameters
                const isUnsafeParameterAssignment = () => {
                    return destinationSignature
                        .getParameters()
                        .some((destinationParameterSymbol, index) => {
                        const destinationParameterType = checker.getTypeOfSymbolAtLocation(destinationParameterSymbol, destinationNode);
                        const sourceParameterSymbol = sourceSignature.getParameters()[index];
                        // eslint-disable-next-line functional/no-conditional-statement
                        if (sourceParameterSymbol === undefined) {
                            return false;
                        }
                        const sourceParameterType = checker.getTypeOfSymbolAtLocation(sourceParameterSymbol, sourceNode);
                        // Source and destination swapped here - function parameters are contravariant.
                        return isUnsafeAssignment(sourceNode, destinationNode, sourceParameterType, destinationParameterType, checker, nextSeenTypesWithPair);
                    });
                };
                const sourceReturnType = sourceSignature.getReturnType();
                const destinationReturnType = destinationSignature.getReturnType();
                // This is an unsafe assignment if...
                return (
                // we're not in an infinitely recursive type,
                seenTypes.every((t) => t.destinationType !== destinationType &&
                    t.sourceType !== sourceType) &&
                    // and the types we're assigning from and to are different,
                    destinationType !== sourceType &&
                    // and the return types of the functions are unsafe assignment,
                    (isUnsafeAssignment(destinationNode, sourceNode, destinationReturnType, sourceReturnType, checker, nextSeenTypesWithPair) ||
                        // or the parameter types of the functions are unsafe assignment.
                        isUnsafeParameterAssignment()));
            });
        });
        const inUnsafeObjectAssignment = (objectTypePairs) => objectTypePairs.some(({ sourceType, destinationType }) => {
            // TODO https://github.com/danielnixon/eslint-plugin-total-functions/issues/100
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            const nextSeenTypesWithPair = nextSeenTypes.concat({
                destinationType,
                sourceType,
            });
            // This is an unsafe assignment if...
            return (
            // we're not in an infinitely recursive type,
            seenTypes.every((t) => t.destinationType !== destinationType &&
                t.sourceType !== sourceType) &&
                // and the types we're assigning from and to are different,
                // TODO this seems to be required to prevent a hang in https://github.com/oaf-project/oaf-react-router
                // Need to work out why and formulate a test to reproduce
                destinationType !== sourceType &&
                // and we're either:
                // unsafe string index assignment, or
                (isUnsafeIndexAssignment(typescript_1.IndexKind.String, destinationNode, sourceNode, destinationType, sourceType, checker, nextSeenTypesWithPair) ||
                    // unsafe number index assignment, or
                    isUnsafeIndexAssignment(typescript_1.IndexKind.Number, destinationNode, sourceNode, destinationType, sourceType, checker, nextSeenTypesWithPair) ||
                    // unsafe property assignment.
                    isUnsafePropertyAssignment(destinationNode, sourceNode, destinationType, sourceType, checker, nextSeenTypesWithPair)));
        });
        return (inUnsafeObjectAssignment(objectTypePairs) ||
            isUnsafeFunctionAssignment(typePairs));
    };
    return {
        // eslint-disable-next-line functional/no-return-void
        TSTypeAssertion: (node) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (node.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                node.typeAnnotation.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                node.typeAnnotation.typeName.name === "const") {
                // Always allow `as const`.
                return;
            }
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = destinationNode.expression;
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringTSTypeAssertion",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        TSAsExpression: (node) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (node.typeAnnotation.type === experimental_utils_1.AST_NODE_TYPES.TSTypeReference &&
                node.typeAnnotation.typeName.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                node.typeAnnotation.typeName.name === "const") {
                // Always allow `as const`.
                return;
            }
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = destinationNode.expression;
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringTSAsExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        VariableDeclaration: (node) => {
            // eslint-disable-next-line functional/no-expression-statement, @typescript-eslint/prefer-readonly-parameter-types, functional/no-return-void
            node.declarations.forEach((declaration) => {
                // eslint-disable-next-line functional/no-conditional-statement
                if (declaration.id.type === experimental_utils_1.AST_NODE_TYPES.Identifier &&
                    declaration.id.typeAnnotation === undefined) {
                    // If there is no type annotation then there's no risk of unsafe assignment.
                    return;
                }
                // eslint-disable-next-line functional/no-conditional-statement
                if (declaration.init === null) {
                    // If there is no type annotation then there's no risk of unsafe assignment.
                    return;
                }
                const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(declaration.id);
                const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(declaration.init);
                const leftType = checker.getTypeAtLocation(leftTsNode);
                const rightType = checker.getTypeAtLocation(rightTsNode);
                // eslint-disable-next-line functional/no-conditional-statement
                if (isUnsafeAssignment(leftTsNode, rightTsNode, leftType, rightType, checker)) {
                    // eslint-disable-next-line functional/no-expression-statement
                    context.report({
                        node: node,
                        messageId: "errorStringVariableDeclaration",
                    });
                }
            });
        },
        // eslint-disable-next-line functional/no-return-void
        AssignmentExpression: (node) => {
            const leftTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.left);
            const rightTsNode = parserServices.esTreeNodeToTSNodeMap.get(node.right);
            const leftType = checker.getTypeAtLocation(leftTsNode);
            const rightType = checker.getTypeAtLocation(rightTsNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafeAssignment(leftTsNode, rightTsNode, leftType, rightType, checker)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringAssignmentExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        ReturnStatement: (node) => {
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            // eslint-disable-next-line functional/no-conditional-statement
            if (tsNode.expression === undefined) {
                return;
            }
            const destinationType = checker.getContextualType(tsNode.expression);
            const sourceType = checker.getTypeAtLocation(tsNode.expression);
            // eslint-disable-next-line functional/no-conditional-statement
            if (destinationType !== undefined &&
                isUnsafeAssignment(tsNode.expression, tsNode.expression, destinationType, sourceType, checker)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node,
                    messageId: "errorStringArrowFunctionExpression",
                });
            }
        },
        // TODO: YieldExpression?
        // eslint-disable-next-line functional/no-return-void
        ArrowFunctionExpression: (node) => {
            // eslint-disable-next-line functional/no-conditional-statement
            if (node.returnType === undefined) {
                return;
            }
            // TODO fix this
            // eslint-disable-next-line total-functions/no-unsafe-mutable-readonly-assignment
            const destinationNode = parserServices.esTreeNodeToTSNodeMap.get(node.returnType.typeAnnotation);
            const destinationType = checker.getTypeAtLocation(destinationNode);
            const sourceNode = parserServices.esTreeNodeToTSNodeMap.get(node.body);
            const sourceType = checker.getTypeAtLocation(sourceNode);
            // eslint-disable-next-line functional/no-conditional-statement
            if (isUnsafeAssignment(destinationNode, sourceNode, destinationType, sourceType, checker)) {
                // eslint-disable-next-line functional/no-expression-statement
                context.report({
                    node: node.body,
                    messageId: "errorStringArrowFunctionExpression",
                });
            }
        },
        // eslint-disable-next-line functional/no-return-void
        CallExpression: (node) => {
            const tsNode = parserServices.esTreeNodeToTSNodeMap.get(node);
            // eslint-disable-next-line functional/no-expression-statement, @typescript-eslint/prefer-readonly-parameter-types, functional/no-return-void
            tsNode.arguments.forEach((argument, i) => {
                var _a;
                const argumentType = checker.getTypeAtLocation(argument);
                const paramType = checker.getContextualType(argument);
                // eslint-disable-next-line functional/no-conditional-statement
                if (paramType !== undefined &&
                    isUnsafeAssignment(argument, argument, paramType, argumentType, checker)) {
                    // eslint-disable-next-line functional/no-expression-statement
                    context.report({
                        node: (_a = node.arguments[i]) !== null && _a !== void 0 ? _a : node,
                        messageId: "errorStringCallExpression",
                    });
                }
            });
        },
    };
};
exports.createNoUnsafeAssignmentRule = createNoUnsafeAssignmentRule;
