import { RuleModule } from "@typescript-eslint/experimental-utils/dist/ts-eslint";
import { ESLintUtils } from "@typescript-eslint/experimental-utils";

/**
 * An ESLint rule to enforce TypeScript strict mode.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const requireStrictMode: RuleModule<
  | "strict"
  | "noUncheckedIndexedAccess"
  | "useUnknownInCatchVariables"
  | "strictFunctionTypes"
  | "strictBindCallApply"
  | "strictNullChecks"
  | "strictPropertyInitialization",
  readonly []
> = {
  meta: {
    type: "problem",
    docs: {
      description: "Enforces the use of TypeScript's strict mode.",
      recommended: "error",
      url: "https://github.com/danielnixon/eslint-plugin-total-functions",
    },
    messages: {
      strict: "TypeScript's strict mode is required.",
      noUncheckedIndexedAccess:
        "TypeScript's noUncheckedIndexedAccess mode is required.",
      useUnknownInCatchVariables:
        "Do not disable the useUnknownInCatchVariables compiler option.",
      strictFunctionTypes:
        "Do not disable the strictFunctionTypes compiler option.",
      strictBindCallApply:
        "Do not disable the strictBindCallApply compiler option.",
      strictNullChecks: "Do not disable the strictNullChecks compiler option.",
      strictPropertyInitialization:
        "Do not disable the strictPropertyInitialization compiler option.",
    },
    schema: [],
  },
  create: (context) => {
    const parserServices = ESLintUtils.getParserServices(context);
    const options = parserServices.program.getCompilerOptions();

    return {
      // eslint-disable-next-line functional/no-return-void
      Program: (node) => {
        const mustBeEnabled = ["strict", "noUncheckedIndexedAccess"] as const;
        const mustNotBeDisabled = [
          "strictFunctionTypes",
          "strictBindCallApply",
          "strictNullChecks",
          "strictPropertyInitialization",
          "useUnknownInCatchVariables",
        ] as const;

        // eslint-disable-next-line functional/no-expression-statement, @typescript-eslint/prefer-readonly-parameter-types, functional/no-return-void
        mustBeEnabled.forEach((option) => {
          // eslint-disable-next-line functional/no-conditional-statement
          if (options[option] !== true) {
            // eslint-disable-next-line functional/no-expression-statement
            context.report({
              node: node,
              messageId: option,
            } as const);
          }
        });

        // eslint-disable-next-line functional/no-expression-statement, @typescript-eslint/prefer-readonly-parameter-types, functional/no-return-void
        mustNotBeDisabled.forEach((option) => {
          // eslint-disable-next-line functional/no-conditional-statement
          if (options[option] === false) {
            // eslint-disable-next-line functional/no-expression-statement
            context.report({
              node: node,
              messageId: option,
            } as const);
          }
        });
      },
    };
  },
} as const;

export default requireStrictMode;
